use poem::middleware::Cors;
use poem::middleware::{AddDataEndpoint, CorsEndpoint};
use poem::{EndpointExt, Route};
use poem_openapi::OpenApiService;

use crate::{
    route::{Api, HealthApi, VersionApi},
    settings::Settings,
};

#[must_use]
pub fn get_connection_pool(settings: &crate::settings::Database) -> sqlx::postgres::PgPool {
    tracing::event!(
        target: "startup",
        tracing::Level::INFO,
        "connecting to database with configuration {:?}",
        settings.clone()
    );
    sqlx::postgres::PgPoolOptions::new()
        .acquire_timeout(std::time::Duration::from_secs(2))
        .connect_lazy_with(settings.get_connect_options())
}

type Server = poem::Server<poem::listener::TcpListener<String>, std::convert::Infallible>;
pub type App = AddDataEndpoint<AddDataEndpoint<CorsEndpoint<Route>, sqlx::PgPool>, Settings>;

pub struct Application {
    server: Server,
    app: poem::Route,
    port: u16,
    database: sqlx::postgres::PgPool,
    settings: Settings,
}

pub struct RunnableApplication {
    server: Server,
    app: App,
}

impl RunnableApplication {
    /// Runs the application until it decides to stop by itself.
    ///
    /// # Errors
    ///
    /// If the server encounters an internal error it cannot recover
    /// from, it will forward it to this function which will forward
    /// it to its caller.
    pub async fn run(self) -> Result<(), std::io::Error> {
        self.server.run(self.app).await
    }
}

impl From<RunnableApplication> for App {
    fn from(value: RunnableApplication) -> Self {
        value.app
    }
}

impl From<Application> for RunnableApplication {
    fn from(val: Application) -> Self {
        let app = val
            .app
            .with(Cors::new())
            .data(val.database)
            .data(val.settings);
        let server = val.server;
        Self { server, app }
    }
}

impl Application {
    async fn setup_db(
        settings: &Settings,
        test_pool: Option<sqlx::postgres::PgPool>,
    ) -> sqlx::postgres::PgPool {
        let database_pool =
            test_pool.map_or_else(|| get_connection_pool(&settings.database), |pool| pool);
        if !cfg!(test) {
            migrate_database(&database_pool).await;
        }
        database_pool
    }

    fn setup_app(settings: &Settings) -> poem::Route {
        let api_service = OpenApiService::new(
            (Api, HealthApi, VersionApi),
            settings.application.clone().name,
            settings.application.clone().version,
        );
        let ui = api_service.swagger_ui();
        poem::Route::new().nest("/", api_service).nest("/docs", ui)
    }

    fn setup_server(
        settings: &Settings,
        tcp_listener: Option<poem::listener::TcpListener<String>>,
    ) -> Server {
        let tcp_listener = tcp_listener.unwrap_or_else(|| {
            let address = format!(
                "{}:{}",
                settings.application.host, settings.application.port
            );
            poem::listener::TcpListener::bind(address)
        });
        poem::Server::new(tcp_listener)
    }
    pub async fn build(
        settings: Settings,
        test_pool: Option<sqlx::postgres::PgPool>,
        tcp_listener: Option<poem::listener::TcpListener<String>>,
    ) -> Self {
        let database_pool = Self::setup_db(&settings, test_pool).await;
        let port = settings.application.port;
        let app = Self::setup_app(&settings);
        let server = Self::setup_server(&settings, tcp_listener);
        Self {
            server,
            app,
            port,
            database: database_pool,
            settings,
        }
    }

    /// Make the app runnable.
    #[must_use]
    pub fn make_app(self) -> RunnableApplication {
        self.into()
    }

    #[must_use]
    pub const fn port(&self) -> u16 {
        self.port
    }
}

async fn migrate_database(pool: &sqlx::postgres::PgPool) {
    sqlx::migrate!()
        .run(pool)
        .await
        .expect("Failed to migrate the database");
}
